@@ -2,7 +2,7 @@ package Trace::Mask::Test;
 use strict;
 use warnings;
 
-use Trace::Mask::Util qw/mask_frame mask_call/;
+use Trace::Mask::Util qw/mask_frame mask_line mask_call/;
 use Trace::Mask::Reference qw/trace/;
 
 use Carp qw/croak/;
@@ -15,15 +15,17 @@ our @EXPORT_OK = qw{
     test_stack_hide test_stack_shift test_stack_stop test_stack_no_start
     test_stack_alter test_stack_shift_and_hide test_stack_shift_short
     test_stack_hide_short test_stack_shift_and_alter test_stack_full_combo
+    test_stack_restart
 };
 
-sub NA() { sub() { 1 } }
+sub NA() { \&NA }
 
 sub test_tracer {
     my %params = @_;
     my $convert = delete $params{convert};
     my $trace   = delete $params{trace};
     my $name    = delete $params{name} || 'tracer test';
+    my $type    = delete $params{type} || 'return';
 
     croak "your must provide a 'convert' callback coderef"
         unless $convert && ref($convert) && reftype($convert) eq 'CODE';
@@ -36,8 +38,8 @@ sub test_tracer {
         my @bad;
 
         for my $test (keys %params) {
-            my $sub = $params{$test};
-            $sub = __PACKAGE__->can($sub) if !ref($sub) && $sub =~ m/^test_/ && $sub !~ m/test_tracer/;
+            my $sub;
+            $sub = __PACKAGE__->can($test) if !ref($test) && $test =~ m/^test_/ && $test !~ m/test_tracer/;
             if($sub && ref($sub) && reftype($sub) eq 'CODE') {
                 $tests{$test} = $sub;
             }
@@ -67,22 +69,85 @@ sub test_tracer {
     my $results = {};
     my $expects = {};
     my $ok;
+    my $sig_die = $SIG{__DIE__};
 
     Test::Stream::Plugin::Subtest::subtest_buffered($name => sub {
+        local $SIG{__DIE__} = $sig_die;
         my $sctx = Test::Stream::Context::context();
         $sctx->set_debug($ctx->debug);
         for my $test (sort keys %tests) {
             my $sub = $tests{$test};
-            my $result = $convert->(scalar $sub->($trace));
-            my $expect = $sub->(scalar \&trace);
+            my $result;
+            {
+                my ($ok, $error, $out, $warning);
+                {
+                    local $@;
+                    local $SIG{__WARN__} = $SIG{__WARN__};
+                    local *STDERR = *STDERR;
+
+                    if ($type eq 'warning') {
+                        $SIG{__WARN__} = sub {
+                            die "Got multiple warnings!\n$warning\n$_[0]\n" if $warning;
+                            ($warning) = @_;
+                        };
+                    }
+                    elsif($type eq 'sigwarn') {
+                        local *Trace::Mask::Test::FAKE;
+                        *STDERR = *Trace::Mask::Test::FAKE;
+                        $warning = "";
+                        open(STDERR, '>', \$warning) || do {
+                            print "Could not open temp STDERR: $!\n";
+                            die "Could not open temp STDERR: $!";
+                        };
+                    }
+
+                    $ok = eval { $out = $sub->($trace); 1 };
+                    $error = $@;
+                }
+
+                if ($type eq 'exception') {
+                    die "Callback did not throw an exception!\n" if $ok;
+                    $ok = 1;
+                    $result = $error;
+                }
+                elsif ($type eq 'return') {
+                    die "Callback did not return anything!\n" unless $out;
+                    $result = $out;
+                }
+                elsif($type eq 'warning' || $type eq 'sigwarn') {
+                    die "Callback did not issue any warnings!\n" unless $warning;
+                    $result = $warning;
+                }
+                else {
+                    die "Invalid type: '$type'\n";
+                }
+
+                die $error unless $ok;
+
+                $result = $convert->($result);
+            }
+            my $expect = $sub->(\&trace);
 
             $results->{$test} = $result;
             $expects->{$test} = $expect;
 
             my $size = min(scalar(@$result), scalar(@$expect));
-            for(my $i = 0; $i < $size; $i++) {
-                for my $j (0, 1) {
-                    delete $expect->[$i]->[$j] unless exists $result->[$i]->[$j];
+            for(my $i = 0; $i < $size; $i++) { # Frame
+                delete $expect->[$i]->[2]; # Remove the mask
+
+                # Args may not be available
+                unless ($result->[$i]->[1] && @{$result->[$i]->[1]}) {
+                    delete $expect->[$i]->[1];
+                    delete $result->[$i]->[1];
+                }
+
+                for (my $j = @{$expect->[$i]->[0]} - 1; $j >= 0; $j--) {
+                    if (exists $result->[$i]->[0]->[$j]) {
+                        $expect->[$i]->[0]->[$j] = sub { 1 } if ref($result->[$i]->[0]->[$j]) && $result->[$i]->[0]->[$j] == \&NA;
+                    }
+                    else {
+                        pop @{$expect->[$i]->[0]};
+                    }
                 }
             }
 
@@ -203,7 +268,7 @@ sub s_and_h_1 { my $code = shift; @_ = (@_); s_and_h_2($code, 'b') }    # line 7
 sub s_and_h_2 { my $code = shift; @_ = (@_); s_and_h_3($code, 'c') }    # line 8
 sub s_and_h_3 { my $code = shift; @_ = (@_); mask_frame(hide  => 1); s_and_h_4($code, 'd') }   # line 9
 sub s_and_h_4 { my $code = shift; @_ = (@_); mask_frame(shift => 1); s_and_h_5($code, 'e') }   # line 10
-sub s_and_h_5 { my $code = shift; @_ = (@_); $code->() }                                       # line 11
+sub s_and_h_5 { my $code = shift; @_ = (@_); $code->() }                                  # line 11
 
 
 
@@ -219,7 +284,7 @@ sub shift_short_1 { my $code = shift; @_ = (@_); shift_short_2($code, 'b') }
 sub shift_short_2 { my $code = shift; @_ = (@_); shift_short_3($code, 'c') }    # line 8
 sub shift_short_3 { my $code = shift; @_ = (@_); shift_short_4($code, 'd') }    # line 9
 sub shift_short_4 { my $code = shift; @_ = (@_); mask_frame(shift => 5); shift_short_5($code, 'e') }    # line 10
-sub shift_short_5 { my $code = shift; @_ = (@_); $code->() }                                            # line 11
+sub shift_short_5 { my $code = shift; @_ = (@_); $code->() }                                       # line 11
 
 
 
@@ -235,7 +300,7 @@ sub hide_short_1 { my $code = shift; @_ = (@_); hide_short_2($code, 'b') }    #
 sub hide_short_2 { my $code = shift; @_ = (@_); hide_short_3($code, 'c') }    # line 8
 sub hide_short_3 { my $code = shift; @_ = (@_); hide_short_4($code, 'd') }    # line 9
 sub hide_short_4 { my $code = shift; @_ = (@_); mask_frame(hide => 5); hide_short_5($code, 'e') }    # line 10
-sub hide_short_5 { my $code = shift; @_ = (@_); $code->() }                                          # line 11
+sub hide_short_5 { my $code = shift; @_ = (@_); $code->() }                                     # line 11
 
 
 
@@ -252,13 +317,13 @@ sub s_and_a_2 { my $code = shift; @_ = (@_); s_and_a_3($code, 'c') }    # line 8
 sub s_and_a_3 {          # line 9
     my $code = shift;    # line 10
     @_ = (@_);           # line 11
-    mask_frame(0 => 'x', 1 => 'x', 2 => 'x', 3 => 'x', 4 => 'x');    # line 12
+    mask_frame(0 => 'x', 1 => 'x', 2 => '100', 3 => 'x', 4 => 'x');    # line 12
     s_and_a_4($code, 'd');                                           # line 13
 }                        # line 14
 sub s_and_a_4 {          # line 15
     my $code = shift;    # line 16
     @_ = (@_);           # line 17
-    mask_frame(0 => 'y', 1 => 'y', 2 => 'y', 3 => 'y', 4 => 'y', shift => 1);    # line 18
+    mask_frame(0 => 'y', 1 => 'y', 2 => '200', 3 => 'y', 4 => 'y', shift => 1);    # line 18
     s_and_a_5($code, 'e');                                                       # line 19
 }                        # line 20
 sub s_and_a_5 { my $code = shift; @_ = (@_); $code->() } # line 21
@@ -274,7 +339,7 @@ sub test_stack_full_combo {              # line 1
 }                                        # line 5
 
 sub full_combo_1 { my $code = shift; @_ = (@_); full_combo_2($code, 'b') }    # line 7
-sub full_combo_2 { my $code = shift; @_ = (@_); full_combo_3($code, 'c') }    # line 8
+sub full_combo_2 { my $code = shift; @_ = (@_); mask_frame('stop' => 1, 'restart' => 1); full_combo_3($code, 'c') }    # line 8
 sub full_combo_3 { my $code = shift; @_ = (@_); full_combo_4($code, 'd') }    # line 9
 sub full_combo_4 { my $code = shift; @_ = (@_); mask_frame('stop' => 1); full_combo_5($code, 'e') }    # line 10
 sub full_combo_5 { my $code = shift; @_ = (@_); full_combo_6($code, 'f') }                             # line 11
@@ -294,6 +359,23 @@ sub full_combo_18 { my $code = shift; @_ = (@_); full_combo_19($code, 's') }
 sub full_combo_19 { my $code = shift; @_ = (@_); full_combo_20($code, 't') }                                                    # line 25
 sub full_combo_20 { my $code = shift; @_ = (@_); mask_frame(no_start => 1); mask_call({no_start => 1}, $code) }                 # line 26
 
+
+
+
+#line 1 "mask_test_restart.pl"
+sub test_stack_restart {                 # line 1
+    my ($callback) = @_;                 # line 2
+    mask_frame(stop => 1, hide => 1);    # line 3
+    restart_1($callback, 'a');           # line 4
+}                                        # line 5
+
+sub restart_1 { my $code = shift; @_ = (@_); restart_2($code, 'b') }                                # line 7
+sub restart_2 { my $code = shift; @_ = (@_); mask_frame('restart' => 1); restart_3($code, 'c') }    # line 8
+sub restart_3 { my $code = shift; @_ = (@_); restart_4($code, 'd') }                                # line 9
+sub restart_4 { my $code = shift; @_ = (@_); mask_frame('stop' => 1); restart_5($code, 'e') }       # line 10
+sub restart_5 { my $code = shift; @_ = (@_); restart_6($code, 'f') }                                # line 11
+sub restart_6 { my $code = shift; @_ = (@_); mask_call({no_start => 1}, $code) }                    # line 12
+
 1;
 
 =pod
@@ -312,13 +394,14 @@ a tracer complies with the specifications.
     use Trace::Mask::Test qw/test_tracer/;
 
     test_tracer(
+        name => 'my tracer',
         trace => \&trace,
+        type  => 'return',
         convert => sub {
             my $stack = shift;
             ...
             return $stack;
         },
-        name => 'my tracer',
     );
 
 =head1 EXPORTS
@@ -346,6 +429,10 @@ manually compare them.
 
 =over 4
 
+=item name => "..."
+
+Specify a name for your test.
+
 =item trace => \&trace
 
 This should be your tracer, or a subroutine that calls it. This subroutine is
@@ -374,9 +461,32 @@ returned from C<NA()> in place of any value that cannot be obtained from your
 stack trace results. In addition it only checks values you have specified, if
 you only list the first 4 fields from caller then only the first 4 are checked.
 
-=item name => "..."
+=item type => 'return'
 
-Specify a name for your test.
+The trace function will return a trace in a scalar, use that as our trace.
+
+=item type => 'exception'
+
+The trace function will throw an exception, intercept it and use the exception
+as the trace.
+
+=item type => 'warning'
+
+The trace function will issue a warning, intecept it using a C<$SIG{__WARN__}>
+override, then use the warning as our trace.
+
+B<Note:> If your tracer issues more than 1 warning and exception will be
+thrown.
+
+B<Note:> Since this uses a C<$SIG{__WARN__}>, it cannot be used to check traces
+that require a custom C<$SIG{__WARN__} override. See the 'sigwarn' type below
+if you are testing a tool that rewrites warnings.
+
+=item type => 'sigwarn'
+
+The trace function will issue a warning, but a custom $SIG{__WARN__} needs to
+modify it before we see it. This will NOT override C<$SIG{__WARN__}>, instead
+it will intercept all output to STDERR when it calls your tracer.
 
 =item %tests
 
@@ -385,6 +495,7 @@ subset of tests then you can list them with a true value.
 
     test_tracer(
         name    => "foo",
+        type    => 'return',
         trace   => \&trace,
         convert => sub { ... },
 
